function make_callback(rhs::RHS, bdryconds::BoundaryConditions)
  cbfn_equation_step  = invoke_after(rhs, rhs.equation)
  cbfn_equation(u, t) = callback_equation(u, t, rhs, isperiodic(bdryconds), rhs.equation)
  cb_equation         = FunctionCallback(cbfn_equation,
                                         CallbackTiming(every_iteration=1,every_dt=0))
  cbfn_tci_step       = invoke_after(rhs, rhs.tci)
  cbfn_tci(u, t)      = callback_tci(u, t, rhs, isperiodic(bdryconds), rhs.tci)
  cb_tci              = FunctionCallback(cbfn_tci,
                                         CallbackTiming(every_iteration=1, every_dt=0))
  cbfn_hrsc_step      = invoke_after(rhs, rhs.hrsc)
  cbfn_hrsc(u, t)     = callback_hrsc(u, t, rhs, isperiodic(bdryconds), rhs.hrsc)
  cb_hrsc = if cbfn_hrsc_step === :substep
    FunctionCallback(cbfn_hrsc, CallbackTiming(force=true))
  else
    FunctionCallback(cbfn_hrsc, CallbackTiming(every_iteration=1,every_dt=0))
  end

  cbs_step = [ (cb_equation, cbfn_equation_step),
               (cb_tci, cbfn_tci_step),
               (cb_hrsc, cbfn_hrsc_step) ]
  callbackset_fullstep = CallbackSet()
  callbackset_substep = CallbackSet()
  for (cb, step) in cbs_step
    step === :fullstep && push!(callbackset_fullstep, cb)
    step === :substep && push!(callbackset_substep, cb)
  end

  return callbackset_fullstep, callbackset_substep
end


# as a default run everything after each full time step
invoke_after(rhs, dispatch) = :fullstep


#######################################################################
#                              Equation                               #
#######################################################################


function callback_equation(state_u, state_t, rhs, isperiodic, eq::EulerEquation)
  @unpack cache, equation = rhs
  wrap_dynamic_variables!(cache, state_u)
  @unpack Sm1, S, flx_S, flx_Sm1 = get_static_variables(cache)
  @unpack t, tm1 = get_global_variables(cache)
  tm1     .= t
  t       .= state_t
  Sm1     .= S
  flx_Sm1 .= flx_S
  broadcast_volume!(internal_energy, equation, cache)
  broadcast_volume!(pressure, equation, cache)
  broadcast_volume!(entropy, equation, cache)
  broadcast_volume!(entropy_flux, equation, cache)
end


#######################################################################
#                                HRSC                                 #
#######################################################################


function callback_hrsc(state_u, state_t, rhs, isperiodic, hrsc::Nothing) end


callback_hrsc(state_u, state_t, rhs, isperiodic, hrsc::AbstractReconstruction) = nothing


invoke_after(rhs, hrsc::HRSC.BernsteinReconstruction) = :substep
function callback_hrsc(state_u, state_t, rhs, isperiodic, hrsc::HRSC.BernsteinReconstruction)
  @unpack cache = rhs
  @unpack rho, q, E = get_dynamic_variables(cache)
  @unpack flag = get_cell_variables(cache)
  HRSC.reconstruct!(rho,    flag, hrsc, isperiodic=isperiodic,
                    m=1e-10, limit_with_boundaries=false)
  HRSC.reconstruct!(q,    flag, hrsc, isperiodic=isperiodic,
                    limit_with_boundaries=false)
  HRSC.reconstruct!(E,  flag, hrsc, isperiodic=isperiodic,
                    m=1e-10, limit_with_boundaries=false)
end



invoke_after(rhs, hrsc::HRSC.WENOReconstruction) = :substep
function callback_hrsc(state_u, state_t, rhs, isperiodic,
    hrsc::HRSC.WENOReconstruction)

  @unpack cache = rhs
  @unpack rho, q, E = get_dynamic_variables(cache)
  @unpack flag = get_cell_variables(rhs.cache)
  HRSC.reconstruct!(rho,  flag, rhs.hrsc, isperiodic=isperiodic)
  HRSC.reconstruct!(q,    flag, rhs.hrsc, isperiodic=isperiodic)
  HRSC.reconstruct!(E,    flag, rhs.hrsc, isperiodic=isperiodic)
end




function callback_hrsc(state_u, state_t, rhs, isperiodic, hrsc::Maybe{HRSC.AbstractHRSC})
  error("To be implemented!")
  error("not yet implemented for type $(typeof(hrsc))")
end


function callback_hrsc(state_u, state_t, rhs, isperiodic, hrsc::HRSC.AbstractArtificialViscosity)
  @unpack cache, mesh, equation = rhs
  @unpack mu, max_v   = get_cell_variables(cache)
  @unpack v           = get_static_variables(cache)
  @unpack rho         = get_dynamic_variables(cache)
  broadcast_volume!(speed, equation, cache)
  Npts, K = layout(mesh)
  mat_max_v = reshape(view(v, :), (Npts, K))
  for k = 1:K
    max_v[k] = maximum(view(mat_max_v, :, k))
  end
  HRSC.compute_viscosity!(
      mu,
      rho, max_v,
      hrsc)
end


function callback_hrsc(state_u, state_t, rhs, isperiodic, hrsc::HRSC.TCIViscosity)
  @unpack cache, mesh = rhs
  @unpack mu, flag    = get_cell_variables(cache)
  HRSC.compute_viscosity!(mu, flag, hrsc)
end


function callback_hrsc(state_u, state_t, rhs, isperiodic, hrsc::HRSC.SmoothedArtificialViscosity)
  callback_hrsc(state_u, state_t, rhs, isperiodic, hrsc.av)
  @unpack cache       = rhs
  @unpack mu          = get_cell_variables(cache)
  @unpack smoothed_mu = get_static_variables(cache)
  smoothed_mu .= vec(hrsc.smoother(mu, isperiodic))
end


function callback_hrsc(state_u, state_t, rhs, isperiodic, hrsc::HRSC.EntropyViscosity)
  @unpack cache, mesh                = rhs
  @unpack rho                        = get_dynamic_variables(cache)
  @unpack S, Sm1, flx_S, flx_Sm1     = get_static_variables(cache)
  @unpack mu, max_v                  = get_cell_variables(cache)
  @unpack t, tm1                     = get_global_variables(cache)
  HRSC.compute_viscosity!(
      mu, 
      rho, max_v, S, Sm1, flx_S, flx_Sm1, t[1], tm1[1],
      hrsc)
end


#######################################################################
#                                 TCI                                 #
#######################################################################


function callback_tci(state_u, state_t, rhs, isperiodic, tci::Nothing) end


function callback_tci(state_u, state_t, rhs, isperiodic, tci::TCI.Threshold)
  error("Threshold TCI must be combined with an indicator that is supposed to be zero for " *
        "for smooth solutions, but not for shocks. TODO Implement an indicator variable.")
end


function callback_tci(state_u, state_t, rhs, isperiodic, tci::TCI.EntropyProduction)
  @unpack cache, mesh                = rhs
  @unpack S, Sm1, flx_S, flx_Sm1, EP = get_static_variables(cache)
  @unpack flag                       = get_cell_variables(cache)
  @unpack EPjump_lhs, EPjump_rhs     = get_bdry_variables(cache)
  @unpack t, tm1                     = get_global_variables(cache)
  TCI.compute_indicator!(flag, EP, EPjump_lhs, EPjump_rhs,
                         S, Sm1, flx_S, flx_Sm1, t[1], tm1[1],
                         tci)
end


function callback_tci(state_u, state_t, rhs, isperiodic,
    tci::Union{TCI.Diffusion, TCI.Minmod, TCI.Threshold,
               TCI.ModalDecayAverage, TCI.ModalDecayHighest})

  @unpack cache, mesh = rhs
  @unpack rho         = get_dynamic_variables(cache)
  @unpack flag        = get_cell_variables(cache)
  TCI.compute_indicator!(flag, rho, tci)
end
