Base.@kwdef mutable struct RHS{T_Equation <:EulerEquation, 
                               T_RSolver  <:AbstractRiemannSolver,
                               T_HRSC     <:Maybe{HRSC.AbstractHRSC},
                               T_TCI      <:Maybe{TCI.AbstractTCI},
                               T_LDG_RSolver<:Maybe{AbstractRiemannSolver},
                               T_AV_RSolver<:Maybe{AbstractRiemannSolver}}

  cache::Cache
  mesh::Mesh
  equation::T_Equation
  rsolver::T_RSolver
  hrsc::T_HRSC               = nothing
  tci::T_TCI                 = nothing
  ldg_rsolver::T_LDG_RSolver = nothing
  av_rsolver::T_AV_RSolver   = nothing
end


#######################################################################
#                         Register variables                          #
#######################################################################


# Julia limitation/bug: Overloading imported functions without their Module prefix shadows their
# definition in this module without a warning or error. This does not happen for methods
# defined in Base.
# Because of that we prefix with _ below.
# Cf. https://github.com/JuliaLang/julia/issues/15510
function _register_variables!(rhs::RHS)
  register_variables!(rhs.cache,
    dynamic_variablenames = (:rho, :q, :E,),     # density, momentum density, energy
    rhs_variablenames     = (:rhs_rho, :rhs_q, :rhs_E),
    static_variablenames  = (:flx_rho, :flx_q, :flx_E,
                             :p, :eps, :v,        # pressure, internal energy, speed
                             :S, :Sm1,           # entropy
                             :flx_S, :flx_Sm1),  # entropy flux
    cell_variablenames    = (:max_v,),
    global_variablenames  = (:t, :tm1) # recent time stamps
  )
  _register_variables!(rhs, rhs.hrsc)
  _register_variables!(rhs, rhs.tci)
end


_register_variables!(rhs, tci_or_hrsc::Nothing) = nothing

_register_variables!(rhs, hrsc::AbstractReconstruction) = nothing


function _register_variables!(rhs, hrsc::HRSC.AbstractArtificialViscosity)
  register_variables!(rhs.cache,
    static_variablenames  = (:ldg_rho, :ldg_q, :ldg_E,              # local DG variable
                             :flx_ldg_rho, :flx_ldg_q, :flx_ldg_E), # local DG flux
    cell_variablenames    = (:mu,) # 'one viscosity to rule them all'
    )
end


function _register_variables!(rhs, hrsc::HRSC.SmoothedArtificialViscosity)
  _register_variables!(rhs, hrsc.av)
  register_variables!(rhs.cache,
    static_variablenames  = (:smoothed_mu,)
  )
end


function _register_variables!(rhs, tci::TCI.AbstractTCI)
  register_variables!(rhs.cache, cell_variablenames = (:flag,))
end


function _register_variables!(rhs, tci::TCI.EntropyProduction)
  register_variables!(rhs.cache, 
                      cell_variablenames = (:flag,),
                      static_variablenames = (:EP,), # entropy production
                      bdry_variablenames   = (:EPjump_lhs, :EPjump_rhs) 
                                              # entropy production due to jumps at interfaces
  )
end


#######################################################################
#                              Timestep                               #
#######################################################################


function timestep(state_u, t, rhs::RHS)
  wrap_dynamic_variables!(rhs.cache, state_u)
  return timestep(rhs, rhs.hrsc)
end


function timestep(rhs::RHS, hrsc::Maybe{HRSC.AbstractHRSC})
  @unpack cache, mesh, equation = rhs

  broadcast_volume!(speed, equation, cache)
  v = get_variable(cache, :v)
  vmax = dg1d.absolute_maximum(view(v,:))
  vmax_limit = 1e4
  if vmax > vmax_limit
    @warn "Limiting timestep due to maximum speed exceeding $vmax_limit"
    vmax = vmax_limit
  end

  @unpack dl, x, CFL, element = mesh
  @unpack N = element

  dt = CFL * dl / (N * vmax)
  return dt
end


function timestep(rhs::RHS, hrsc::HRSC.AbstractArtificialViscosity)
  @unpack cache, mesh, equation = rhs

  broadcast_volume!(speed, equation, cache)
  v = get_variable(cache, :v)
  vmax = dg1d.absolute_maximum(view(v,:))
  vmax_limit = 1e4
  if vmax > vmax_limit
    @warn "Limiting timestep due to maximum speed exceeding $vmax_limit"
    vmax = vmax_limit
  end
  smoothed_mu = get_variable(cache, :smoothed_mu)
  mumax = dg1d.absolute_maximum(view(smoothed_mu, :))

  @unpack dl, x, CFL, element = mesh
  @unpack N = element

  dt = CFL / (vmax * N / dl + mumax * N^3 / dl^2)
  dt_limit = 1e-7
  if dt < dt_limit
    @warn "Limiting timestep due to dt being smaller than $dt_limit"
    dt = dt_limit
  end
  return dt
end


#######################################################################
#                           RHS computation                           #
#######################################################################


function rhs!(state_du, state_u, t, rhs::RHS,
    bdryconds::BoundaryConditions, ldg_bdryconds::Maybe{BoundaryConditions},
    av_bdryconds::Maybe{BoundaryConditions})

  @unpack cache = rhs
  wrap_dynamic_variables!(cache, state_u)
  wrap_rhs_variables!(cache, state_du)
  rhs!(rhs, t, rhs.hrsc, bdryconds, ldg_bdryconds, av_bdryconds)
end


function rhs!(rhs::RHS, t, hrsc::Nothing, bdryconds, ldg_bdryconds, av_bdryconds)

  @unpack cache, mesh, equation, rsolver = rhs
  @unpack rho, q, E                     = get_dynamic_variables(cache)
  @unpack flx_rho, flx_q, flx_E, p, eps = get_static_variables(cache)
  @unpack rhs_rho, rhs_q, rhs_E         = get_rhs_variables(cache)
  @unpack lhs_numflx_rho, rhs_numflx_rho,
          lhs_numflx_q  , rhs_numflx_q  ,
          lhs_numflx_E  , rhs_numflx_E  = get_bdry_variables(cache)

  broadcast_volume!(flux, equation, cache)
  broadcast_faces!(lax_friedrich_flux, rsolver, cache, mesh)
  broadcast_boundaryconditions!(lax_friedrich_flux, bdryconds, cache, mesh, t)

  compute_rhs_weak_form!(rhs_rho, flx_rho,  lhs_numflx_rho, rhs_numflx_rho, mesh)
  compute_rhs_weak_form!(rhs_q,   flx_q,    lhs_numflx_q,   rhs_numflx_q,   mesh)
  compute_rhs_weak_form!(rhs_E,   flx_E,    lhs_numflx_E,   rhs_numflx_E,   mesh)

  return
end


function rhs!(rhs::RHS, t, hrsc::AbstractReconstruction, bdryconds, ldg_bdryconds, av_bdryconds)

  @unpack cache, mesh, equation, rsolver = rhs
  @unpack rho, q, E                      = get_dynamic_variables(cache)
  @unpack flx_rho, flx_q, flx_E, p, eps  = get_static_variables(cache)
  @unpack flag                           = get_cell_variables(cache)
  @unpack rhs_rho, rhs_q, rhs_E          = get_rhs_variables(cache)
  @unpack lhs_numflx_rho, rhs_numflx_rho,
          lhs_numflx_q  , rhs_numflx_q  ,
          lhs_numflx_E  , rhs_numflx_E   = get_bdry_variables(cache)

  broadcast_volume!(flux, equation, cache)
  broadcast_faces!(lax_friedrich_flux, rsolver, cache, mesh)
  broadcast_boundaryconditions!(lax_friedrich_flux, bdryconds, cache, mesh, t)

  compute_rhs_weak_form!(rhs_rho, flx_rho,  lhs_numflx_rho, rhs_numflx_rho, mesh)
  compute_rhs_weak_form!(rhs_q,   flx_q,    lhs_numflx_q,   rhs_numflx_q,   mesh)
  compute_rhs_weak_form!(rhs_E,   flx_E,    lhs_numflx_E,   rhs_numflx_E,   mesh)

  return
end


function rhs!(rhs::RHS, t, hrsc::HRSC.AbstractArtificialViscosity,
    bdryconds, ldg_bdryconds, av_bdryconds)

  @unpack cache, mesh, equation, rsolver, ldg_rsolver, av_rsolver = rhs

  @unpack rho, q, E                         = get_dynamic_variables(cache)
  @unpack flx_rho, flx_q, flx_E, p, eps,
          ldg_rho, ldg_q, ldg_E,
          flx_ldg_rho, flx_ldg_q, flx_ldg_E = get_static_variables(cache)
  @unpack rhs_rho, rhs_q, rhs_E             = get_rhs_variables(cache)
  @unpack lhs_numflx_rho,     rhs_numflx_rho,
          lhs_numflx_q,       rhs_numflx_q,
          lhs_numflx_E,       rhs_numflx_E,
          lhs_numflx_ldg_rho, rhs_numflx_ldg_rho,
          lhs_numflx_ldg_q,   rhs_numflx_ldg_q,
          lhs_numflx_ldg_E,   rhs_numflx_ldg_E  = get_bdry_variables(cache)

  ## solve auxiliary equation: q + ∂x u = 0
  broadcast_volume!(ldg_flux, equation, cache)
  broadcast_faces!(central_flux, ldg_rsolver, cache, mesh)
  broadcast_boundaryconditions!(central_flux, ldg_bdryconds, cache, mesh, t)
  compute_rhs_weak_form!(ldg_rho, flx_ldg_rho,  lhs_numflx_ldg_rho, rhs_numflx_ldg_rho, mesh)
  compute_rhs_weak_form!(ldg_q,   flx_ldg_q,    lhs_numflx_ldg_q,   rhs_numflx_ldg_q,   mesh)
  compute_rhs_weak_form!(ldg_E,   flx_ldg_E,    lhs_numflx_ldg_E,   rhs_numflx_ldg_E,   mesh)

  ## compute rhs of regularized equation: ∂t u + ∂x f + ∂x μ q = 0
  broadcast_volume!(av_flux, equation, cache)
  broadcast_faces!(mean_flux, av_rsolver, cache, mesh)
  broadcast_boundaryconditions!(mean_flux, av_bdryconds, cache, mesh, t)
  broadcast_volume!(flux, equation, cache)
  broadcast_faces!(lax_friedrich_flux, rsolver, cache, mesh)
  broadcast_boundaryconditions!(lax_friedrich_flux, bdryconds, cache, mesh, t)

  @. flx_rho        += flx_ldg_rho
  @. flx_q          += flx_ldg_q
  @. flx_E          += flx_ldg_E
  @. lhs_numflx_rho += lhs_numflx_ldg_rho
  @. lhs_numflx_q   += lhs_numflx_ldg_q
  @. lhs_numflx_E   += lhs_numflx_ldg_E
  @. rhs_numflx_rho += rhs_numflx_ldg_rho
  @. rhs_numflx_q   += rhs_numflx_ldg_q
  @. rhs_numflx_E   += rhs_numflx_ldg_E

  compute_rhs_weak_form!(rhs_rho, flx_rho,  lhs_numflx_rho, rhs_numflx_rho, mesh)
  compute_rhs_weak_form!(rhs_q,   flx_q,    lhs_numflx_q,   rhs_numflx_q,   mesh)
  compute_rhs_weak_form!(rhs_E,   flx_E,    lhs_numflx_E,   rhs_numflx_E,   mesh)

  return
end
