struct InitialData{T} end


function initialdata!(rhs::RHS, parameters)
  type = query(parameters, "type")
  type = Symbol(lowercase(replace(type, " "=>"_")))
  id_type = InitialData{type}()

  initialdata_equation!(id_type, rhs, parameters)
  return
  # initialdata_entropy_variables!(id_type, rhs, parameters)
  initialdata_hrsc!(id_type, rhs, rhs.hrsc, parameters)
  initialdata_tci!(id_type, rhs, rhs.tci, parameters)
end


#######################################################################
#                     Initial data gravity sector                     #
#######################################################################


function initialdata_equation!(id::InitialData, rhs, parameters)
  error_not_implemented("initialdata_equation!", typeof(id))
end


function initialdata_equation!(::InitialData{:bondi_accretion}, rhs, parameters)
  TODO(initialdata_equation!)
end


function initialdata_equation!(::InitialData{:tov}, rhs, parameters)

  filename = query(parameters, "filename")

  # read TOV result from file
  # only contains solution in star interior
  output_dir = dg1d.get_output_dir()
  data  = TOV.read_data(joinpath(output_dir, "TOV.h5"))
  ϕint  = data.ϕ
  Rint  = data.R
  rint  = data.r
  pint  = data.p
  K     = data.K
  Γ     = data.Γ
  M     = data.M
  Rsurf = Rint[end]

  # interpolate data onto our grid
  # we don't assume a special grid layout for the loaded data

  @unpack mesh = rhs
  @unpack Npts = mesh.element
  intrp_R = vec(mesh.x) # DG coordinates
  NN = length(intrp_R)

  # determine if grid contains r = 0 point
  R_ismirrored = intrp_R[1] ≈ -intrp_R[end]

  if !R_ismirrored
    idx_origin = findall(Ri -> isapprox(Ri, 0), intrp_R)
    @assert length(idx_origin) > 0 "Non-symmetric grids must include coordinate origin"
  end

  # determine our grid setup
  R_includes_origin = (
    !R_ismirrored # LGL grids always contain bdry points
    || (
      iseven(mesh.K) ||  # origin falls at interface between two cells
      (isodd(mesh.K) && isodd(mesh.element.Npts)) # origin sits in the center
                                                  # of a cell with a point in the center
    )
  )
  R_origin_isdoubled = (
    R_ismirrored && # only happens for mirrored domains
    iseven(mesh.K)  # and origin falls at interface between two cells
  )
  # if r = 0 only covered by one point we have to exclude the boundary
  # point when mirroring/refelcting back
  exclude_origin = !R_origin_isdoubled && R_includes_origin

  # on a mirrored grid we only do interpolation on one half and then
  # mirror/reflect onto the other
  @show NN
  @show R_ismirrored
  @show R_origin_isdoubled
  @show R_includes_origin
  if R_ismirrored
    idx_innermost_R = if R_origin_isdoubled
      Int64(NN/2) + 1
    elseif R_includes_origin
      Int64((NN+1)/2)
    else
      Int64(NN/2) + 1
    end
    intrp_R = intrp_R[idx_innermost_R:end]
  end
  Rmax = maximum(intrp_R) # star's radius

  @info "TOV star surface located at R = $Rsurf"

  #### setup metric functions A, B and derivatives in isotropic coordinates

  # star interior solution
  Aint = exp.(ϕint)
  # B := r / R, needs special treatment of coordinate singularity.
  Bint = [ (ri == 0) ? 1.0 : ri / Ri for (ri, Ri) in zip(rint, Rint) ]
  # use linear interpolation to obtain value at r = R = 0.
  Bint[1] = (Bint[3] - Bint[2]) / (Rint[3] - Rint[2]) * (0.0 - Rint[2]) + Bint[2]

  # derivatives
  dAint = dg1d.FD.central_dx(Aint, Rint)
  dBint = dg1d.FD.central_dx(Bint, Rint)

  # star exterior solution =^= Schwarzschild solution
  Npts_ext = 1000 # number of points used for interpolation of exterior solution
  Rext = collect(LinRange(Rsurf, Rmax, Npts_ext))
  Aext, Bext = schwarzschild_solution_isotropic_coordinates(Rext, M)

  # merge interior and exterior solutions
  # note: exclude surface point in one of the arrays when merging
  R = cat(Rint, Rext[2:end], dims=1)
  A = cat(Aint, Aext[2:end], dims=1)
  B = cat(Bint, Bext[2:end], dims=1)

  # derivatives
  dA = dg1d.FD.central_dx(A, R)
  dB = dg1d.FD.central_dx(B, R)

  intrp_A   = interpolate(intrp_R[:], R, A)
  intrp_B   = interpolate(intrp_R[:], R, B)
  intrp_A_r  = interpolate(intrp_R[:], R, dA)
  intrp_B_r  = interpolate(intrp_R[:], R, dB)

  if R_ismirrored
    intrp_A    = mirror(intrp_A, exclude_origin)
    intrp_B    = mirror(intrp_B, exclude_origin)
    intrp_A_r  = reflect(intrp_A_r, exclude_origin)
    intrp_B_r  = reflect(intrp_B_r, exclude_origin)
  end

  @assert all(map(X->all(isfinite.(X)), [intrp_A, intrp_B, intrp_A_r, intrp_B_r])) == true ""
    "Initial data interpolation failed!"

  @unpack A, B, A_r, B_r = get_static_variables(rhs.cache)
  A   .= intrp_A
  B   .= intrp_B
  A_r .= intrp_A_r
  B_r .= intrp_B_r

  ### setup hydrodynamic variables
  @unpack equation = rhs
  @unpack atmosphere = equation
  eos = Polytrope(K=K, Gamma=Γ)

  # apply atmosphere treatment
  pmax = maximum(pint)
  ρmax = eos(Density, Pressure, pmax)
  ρatm = atmosphere.factor * ρmax
  patm = eos(Pressure, Density, ρatm)
  pint[@.(pint < 0)] .= 0.0 # just to be sure, they will be replace with patm below anyways
  ρint = @. eos(Density, Pressure, pint)
  idxatm_treatment = @. ρint < ρatm * atmosphere.threshold
  pint[idxatm_treatment] .= patm
  pint[idxatm_treatment] .= ρatm
  vint = zeros(size(Rint))
  ϵint = @. eos(InternalEnergy, Density, ρint)

  # star exterior data
  ρext = ρatm * ones(size(Rext))
  pext = @. eos(Pressure, Density, ρext)
  vext = zeros(size(Rext))
  ϵext = @. eos(InternalEnergy, Density, ρext)

  # merge interior and exterior solutions
  p  = cat(pint, pext[2:end], dims=1)
  ρ  = cat(ρint, ρext[2:end], dims=1)
  vr = cat(vint, vext[2:end], dims=1)
  ϵ  = cat(ϵint, ϵext[2:end], dims=1)

  intrp_p  = interpolate(intrp_R[:], R, p)
  intrp_ρ  = interpolate(intrp_R[:], R, ρ)
  intrp_vr = interpolate(intrp_R[:], R, vr)
  intrp_ϵ  = interpolate(intrp_R[:], R, ϵ)

  if R_ismirrored
    intrp_p  = mirror(intrp_p, exclude_origin)
    intrp_ρ  = mirror(intrp_ρ, exclude_origin)
    intrp_vr = reflect(intrp_vr, exclude_origin)
    intrp_ϵ  = mirror(intrp_ϵ, exclude_origin)
  end

  @unpack p, vr, rho, eps = get_static_variables(rhs.cache)
  @unpack sD, sSr, stau   = get_dynamic_variables(rhs.cache)
  @unpack D, Sr, tau,     = get_static_variables(rhs.cache)

  p .= intrp_p
  rho .= intrp_ρ
  vr .= intrp_vr
  eps .= intrp_ϵ

  broadcast_volume!(prim2cons, equation, rhs.cache)
  broadcast_volume!(cons2scaledcons, equation, rhs.cache)

  @assert all(map(X->all(isfinite.(X)),
    [p, vr, rho, eps, D, Sr, tau, sD, sSr, stau])) == true "Initial data interpolation failed!"

  return
end


function schwarzschild_solution_isotropic_coordinates(R, M)
  Rmin = minimum(R)
  @assert 2 * Rmin >= M
  A = @. (1 - 0.5 * M / R) / (1 + 0.5 * M / R)
  B = @. (1 + 0.5 * M / R)^2
  return A, B
end


function interpolate(x, x_ref, y_ref)
  x_intrp = LinearInterpolation(x_ref, y_ref, extrapolation_bc=Line())
  y = [ x_intrp(xi) for xi in x ]
  return y
end


function mirror(A, exclude_bdry::Bool)
  if exclude_bdry
    return cat(A[end:-1:2], A, dims=1)
  else
    return cat(A[end:-1:1], A, dims=1)
  end
end


function reflect(A, exclude_bdry::Bool)
  if exclude_bdry
    return cat(-A[end:-1:2], A, dims=1)
  else
    return cat(-A[end:-1:1], A, dims=1)
  end
end


function initialdata_equation!(::InitialData{:atmosphere}, rhs, parameters)

  @unpack mesh, equation = rhs
  @unpack Npts = mesh.element
  @unpack eos = equation
  @unpack A, B, A_r, B_r = get_static_variables(rhs.cache)
  R = vec(mesh.x)

  # minkowski metric
  A   .= ones(size(R))
  B   .= ones(size(R))
  A_r .= zeros(size(R))
  B_r .= zeros(size(R))

  # low density atmosphere
  rho_atm = query(parameters, "density", default=1e-11)
  @assert rho_atm > 0.0
  @assert eos isa Polytrope "Require model for cold atmosphere"

  @unpack rho, vr, p, eps = get_static_variables(rhs.cache)

  NN = length(mesh)
  @. rho = rho_atm * $ones(NN)
  @. p   = eos(Pressure, Density, rho, InternalEnergy, eps)
  @. eps = eos(InternalEnergy, Density, rho, Pressure, p)
  @. vr  = $zeros(NN)

  @assert all(map(X->all(isfinite.(X)), [rho, p, eps, vr])) == true ""
    "Invalid initial data encountered!"

  broadcast_volume!(prim2cons, equation, rhs.cache)
  broadcast_volume!(cons2scaledcons, equation, rhs.cache)

  return
end


#######################################################################
#                    Initialdata entropy variables                    #
#######################################################################


function initialdata_entropy_variables!(::InitialData, rhs, parameters)
  TODO(initialdata_entropy_variables!)
end

#######################################################################
#                          InitialData HRSC                           #
#######################################################################

function initialdata_hrsc!(::InitialData, rhs, hrsc, parameters)
  TODO(initialdata_hrsc!)
end

#######################################################################
#                           InitialData TCI                           #
#######################################################################

function initialdata_tci!(::InitialData, rhs, tci, prms)
  TODO(initialdata_tci!)
end
