@doc """
    EntropyViscosity

Entropy Viscosity model, cf. Yu, Hesthaven, 2019.
"""
Base.@kwdef struct EntropyViscosity <: AbstractArtificialViscosity
  mesh::Mesh
  cmax::Float64 = 1.0
  ce::Float64   = 0.0
end


function parse_parameters(::Type{EntropyViscosity}, prms)
  cmax  = query(Float64, prms, "cmax", default=1.0)
  ce    = query(Float64, prms, "ce", default=0.1)
  @assert cmax > 0
  @assert ce > 0
  @assert cmax >= ce
  return (; cmax, ce)
end


function compute_viscosity!(
    mu,                                    # mutated outputs
    u, v, E, E_prev, F, F_prev, t, t_prev, # inputs
    av::EntropyViscosity)

  @unpack mesh = av
  @unpack element, K, L, dl, x, invjac = mesh
  @unpack D, N = element
  @unpack cmax, ce = av

  dt      = t - t_prev

  # h ... 'characteristic' length of sub cells (Yu, Hesthaven, 2019)
  # consistent with Zingan et al., 2013
  h = L / (K * N)

  E = reshape(E, layout(mesh))
  F = reshape(F, layout(mesh))
  E_prev = reshape(E_prev, layout(mesh))
  F_prev = reshape(F_prev, layout(mesh))

  # entropy residual violation
  # Q: Is it really justified to use no viscosity if we have to compute
  # entropy wrt backwards time step?
  if dt > 0.0
    residual = (E - E_prev) / dt + 0.5 * D * (F + F_prev) .* invjac
  else
    residual = zeros(size(u))
  end
  R = vec(maximum(abs.(residual), dims=1))

  # # entropy jump at interfaces
  # intf_F = dg1d.to_interface(F, element, mesh)
  # # TODO What about bdry conditions?
  jump = zeros(K+1)
  jump[2:K] .= abs.(F[end,1:K-1] - F[1,2:K]) * N / dl
  jump[K] = abs.(F[end,K] - F[1,1]) * N / dl
  # # extract jumps occuring at lhs and rhs of a cell's interface
  jump_lhs = jump[2:end]
  jump_rhs = jump[1:end-1]

  avg_E = dg1d.grid_average(E, mesh)
  norm_E = maximum(@. abs(E - avg_E))
  if isapprox(norm_E, 0.0)
    norm_E = 1.0
  end

  max_R = max.(R, jump_lhs, jump_rhs)
  mu_e = max_R / norm_E * ce * (dl / N)^2

  # upper bound on AV
  maxabs_v = vec(maximum(abs.(v), dims=1))
  mu_max = @. cmax * maxabs_v * dl / N

  # final AV
  @. mu = min(mu_e, mu_max)

  return true
end
