#######################################################################
#                                Type                                 #
#######################################################################


struct SmoothedArtificialViscosity{T_AV<:AbstractArtificialViscosity} <: AbstractArtificialViscosity
  av::T_AV
  smoother::Function
end


#######################################################################
#                             Constructor                             #
#######################################################################


function SmoothedArtificialViscosity(mesh::Mesh, av::AbstractArtificialViscosity; smoother_order=2)
  quadr = mesh.element.quadr
  interp_element = dg1d.SpectralElement(smoother_order, quadr)
  interp_mat = dg1d.lagrange_interp_matrix(mesh.element.z, interp_element.z)
  smoother(u, isperiodic) = smooth_C0(u, interp_mat, isperiodic)
  return SmoothedArtificialViscosity(av, smoother)
end


#######################################################################
#                               Methods                               #
#######################################################################


"""
Smooth cell wise constant values const_v over neighbouring grid cells 
and use a linear polynomial to interpolate nodal values.
Based on Yu (2017), Discacciati (2019), Kloeckner (2011).
const_v ... constant values in each cell
interp_mat ... interpolation matrix
"""
function smooth_C0(const_v, interpmat, isperiodic=true)
  interp_order = size(interpmat, 2) - 1
  if (interp_order != 1 && interp_order != 2)
    error("Input size of interpolation matrix not compatible!")
  end
  avg_v_lhs = 0.5 * (const_v .+ circshift(const_v, 1))
  avg_v_rhs = circshift(avg_v_lhs, -1)
  if !isperiodic
    avg_v_lhs[1]    = const_v[1]
    avg_v_rhs[end]  = const_v[end]
  end # if

  if (interp_order == 2)
    ref_v = hcat(avg_v_lhs, const_v, avg_v_rhs)
  else
    ref_v = hcat(avg_v_lhs, avg_v_rhs)
  end
  # interpolate
  smoothed_v = interpmat * transpose(ref_v)
  return abs.(smoothed_v)
end
