#######################################################################
#                                Types                                #
#######################################################################


struct WENOReconstruction <: AbstractReconstruction
  order::Int64
  threshold::Float64
  mesh::Mesh
  w::Int64
  subcell_bdry_coordinates::Vector{Vector{Float64}} # computational grid ⊂ [-2w+1,2w+1]
  subcell_polynomial_basis::Vector{Matrix{Float64}} # sampled on bdry coordinates
  optimal_weights::Matrix{Float64}
  split_weights::NTuple{2,Matrix{Float64}}
  norm_split_weights::NTuple{2,Vector{Float64}}
  requires_split::Vector{Bool}
  smoothness_estimator::Vector{Matrix{Float64}}

  function WENOReconstruction(; mesh, order, threshold)

    @unpack z = mesh.element

    w = Int64((order-1)/2)

    _, ncells = layout(mesh)
    @assert 0 < w <= ncells "Not enough cells for WENO reconstruction of order $order"

    # build w+1 subcell grids which all overlap with x
    subcell_bdry_coordinates = [ [ 1.0 + 2.0 * (- w + l + k) for l = -1:w ] for k = 0:w ]
    subcell_polynomial_basis = [ bdry_interpolating_polynomial_basis(z, xcell)
                                 for xcell in subcell_bdry_coordinates ]

    opt_weights = optimal_weights(z, w)
    smoothness_estimator = smoothness_estimator_matrix(w)

    split_weights, norm_split_weights, requires_split = split_optimal_weights(opt_weights)

    return new(order, threshold, mesh, w, subcell_bdry_coordinates, subcell_polynomial_basis,
              opt_weights, split_weights, norm_split_weights, requires_split,
              smoothness_estimator)
  end

end


#######################################################################
#                            Constructors                             #
#######################################################################


function parse_parameters(::Type{WENOReconstruction}, prms)
  # query
  order = query(Int64, prms, "order", default = 3)
  threshold = query(Float64, prms, "threshold", default = 1e-6)
  # sanitize
  # TODO Check if isodd is really required
  @assert order > 0 && isodd(order) "'0 < order' and 'order is odd' required"
  @assert threshold > 0 "'threshold > 0' required"
  return (; order, threshold)
end


#######################################################################
#                               Methods                               #
#######################################################################


@doc """
    bdry_interpolating_polynomial_basis(xbase, xbdrys)

Compute nodal representation of the `w`th order polynomial basis `q_n` from the subcell
boundary coordinates `xbdrys`.  The `n`th row holds the nodal coefficients of the `n`th
polynomial `q_n` sampled on points `xbase`.

Based on Bugner 2017 eq. (3.78).
If you multiply the matrix from the right with a vector of average values you obtain the
boundary interpolation polynomial sampled on `xbase`.
"""
function bdry_interpolating_polynomial_basis(xbase, xbdrys)
  npts_bdry = length(xbdrys)
  n_bdry = npts_bdry - 1
  D = dg1d.first_derivative_matrix(xbdrys)
  I = dg1d.lagrange_interp_matrix(xbase, xbdrys)
  ID = I * D
  qs = zeros(n_bdry, length(xbase))
  for n = 0:n_bdry-1
    for m = n:n_bdry-1
      qs[n+1,:] .+= ID[:,m+2]
    end
  end
  qs .*= 2
  return qs
end



@doc """
    optimal_weights(x, w)

Compute optimal WENO weights matrix for a `2w+1` order reconstruction on a grid `x`,
where the weights for different cells are organized row-wise.

The computation follows Bugner 2017, section 3.3.1.
Note that the computation is done numerically in such a way, that all intermediate computations
are carried out including exact polynomial approxiamations.
"""
function optimal_weights(x, w)
  # w+1 subcells which all overlap with x ⊂ [-1,1]
  xsubcells = [ [ 1.0 + 2.0 * (- w + l + k) for l = -1:w ] for k = 0:w ]
  # grid that spans full reconstruction stencile
  xfullcell = [ 1.0 + 2.0 * (- w + l ) for l = -1:2w ]

  ps_basis = [ bdry_interpolating_polynomial_basis(x, xcell) for xcell in xsubcells ]
  q_basis = bdry_interpolating_polynomial_basis(x, xfullcell)

  # for every point in x solve the least squares problem (3.81) from Bugner, p 43.
  npts = length(x)
  weights = zeros(w+1, npts)
  for idx_pt in 1:npts
    # build coefficient matrix from sub-cell-spanning polynomial basis
    A = zeros(2*w+1, w+1)
    for i_cell in 1:w+1
      A[i_cell:i_cell+w,i_cell] .= ps_basis[i_cell][:, idx_pt]
    end
    # build inhomogenity from full-cell-spanning polynomial basis
    b = q_basis[:, idx_pt]
    # solve least-squares problem for weights
    weights[:, idx_pt] .= A \ b
  end

  return weights
end


function split_optimal_weights(optimal_weights)

  theta = 3
  requires_split = [ any(ow_cell .< 0.0) for ow_cell in eachcol(optimal_weights) ]
  split_weights_plus, split_weights_minus = [ zeros(size(optimal_weights)) for _ = 1:2 ]
  # eq (2.1)
  for (i, requires) in enumerate(requires_split)
    !requires && continue
    ow = optimal_weights[:,i]
    split_weights_plus[:,i]  .= @. (ow + theta * abs(ow)) / 2
    split_weights_minus[:,i] .= @. split_weights_plus[:,i] - ow
  end
  # eq (2.2), rescale/normalize
  norm_split_weights_plus  = vec(sum(split_weights_plus, dims=1))
  norm_split_weights_minus = vec(sum(split_weights_minus, dims=1))
  for (i, requires) in enumerate(requires_split)
    !requires && continue
    split_weights_plus[:,i]  ./= norm_split_weights_plus[i]
    split_weights_minus[:,i] ./= norm_split_weights_minus[i]
  end

  return tuple(split_weights_plus, split_weights_minus),
         tuple(norm_split_weights_plus, norm_split_weights_minus),
         requires_split
end



@doc """
    smoothness_estimator_matrix(w)

Compute matrix representation of smoothness estimator for an order `2w+1` WENO subcell splitting.
If `M` is this matrix, then smoothness estimation can be done by computing `u' * M * u`, where
`u` contain the `2w+1` sub cell averages.

Based on Bugner 2017, section 3.3.1.
Note that the computation is done numerically in such a way, that all intermediate computations
are carried out including exact polynomial approxiamations.
"""
function smoothness_estimator_matrix(w)

  # setup a local quadrature rule to evaluate integrals over derivatives of lagrange polynomials
  element = SpectralElement(2*w, :LGL)
  @unpack iproduct, z, Npts = element

  # build w+1 subcell grids which all overlap with x
  xsubcells = [ [ 1.0 + 2.0 * (- w + l + k) for l = -1:w ] for k = 0:w ]
  # compute nodal basis coefficients for each subcell polynomial `p_i(x)`
  # Note that here `p_i` is represented as an order `w+1` polynomial, although it is actually
  # of order `w` because it is the result of differentiating a `w+1` polynomial. We do so
  # in order to reuse the nodal derivative matrices.
  ps_basis = [ bdry_interpolating_polynomial_basis(xcell, xcell) for xcell in xsubcells ]

  # carry out sum of integrals over squared derivatives
  Es = [ zeros(w+1, w+1) for _ in 1:w+1 ] # smoothness estimator matrices for each subcell
  for (n, (xs, ps)) in enumerate(zip(xsubcells, ps_basis))

    # interpolate from subcell grid to quadrature grid
    I = dg1d.lagrange_interp_matrix(z, xs)

    # loop over derivatives
    for k in 1:w

      Dk = dg1d.derivative_matrix(k, xs)
      Lk = I * Dk # kth derivative of lagrange polynomial interpolated to quadrature grid
      integral_L2 = zeros(w+2,w+2) # because the ps' are represented as `w+1`th polynomials

      for i = 1:w+2, j = i:w+2 # collapsed loop because integral_L2 matrix is symmetric
        integral_L2[i,j] = iproduct(Lk[:,i], Lk[:,j])
        (i == j) && continue
        integral_L2[j,i] = integral_L2[i,j]
      end

      Es[n] .+= 2^(2*k-1) * ps * integral_L2 * transpose(ps)
    end

  end

  return Es
end



@doc """
    reconstruct!(u, flags, weno::WENOReconstruction)

Perform WENO reconstruction inplace in `u` for troubled cells indicated by `flags`.
"""
function reconstruct!(u, flags, weno::WENOReconstruction; isperiodic::Bool=false)

  @unpack w = weno
  @unpack K = weno.mesh
  @unpack Npts = weno.mesh.element
  shape = (Npts, K)

  mat_u = reshape(u, shape)
  # mat_utmp = deepcopy(mat_u)

  # cell-averages
  # TODO Move some of these storage variables into WENOReconstruction
  uavg = dg1d.cellwise_average(mat_u, weno.mesh)
  omegas, omegas_plus, omegas_minus = [ zeros(Float64, w+1, Npts) for _ = 1:3 ]
  tmp_omegas = zeros(Float64, w+1, Npts)
  tmp_uweno = zeros(Float64, weno.mesh.element.Npts)
  range = collect(-w:w)
  idx_range = similar(range)

  gammas = weno.optimal_weights
  gammas_plus, gammas_minus = weno.split_weights
  sigmas_plus, sigmas_minus = weno.norm_split_weights

  # go through troubled cells
  for (idx, flag) in enumerate(flags)

    isapprox(flag, 0.0) && continue

    near_bdry = idx <= w || idx >= K - w
    if near_bdry
      continue
      error("WENO reconstruction extends across non-periodic boundary")
    end

    idx_range .= range .+ idx
    if isperiodic && near_bdry
      idx_range .= periodic_index.(idx_range, K)
    end

    uavg_stencile = uavg[idx_range] # full width stencile surrounding troubled cell

    fill!(omegas_plus,  0.0)
    fill!(omegas_minus, 0.0)
    fill!(omegas,       0.0)

    for i = 1:w+1 # for every stencile
      subrange = 1+(i-1):1+(i-1)+w
      uavg_substencile = uavg_stencile[subrange] # covers ith reconstruction element

      # estimate smoothness, (3.69)
      smoothness_i = transpose(uavg_substencile) * weno.smoothness_estimator[i] * uavg_substencile

      # compute non-linear weights, pointwise
      for k = 1:Npts
        if weno.requires_split[k] # due to optimal weights being negative
          # collapsed eq (2.1), (2.2) and (1.4) in Shi et al. 1996
          omegas_plus[i,k]  = gammas_plus[i,k]  / (weno.threshold + smoothness_i)^2
          omegas_minus[i,k] = gammas_minus[i,k] / (weno.threshold + smoothness_i)^2
        else
          # eq (3.70) in bugner's thesis
          omegas[i,k] = gammas[i,k] / (weno.threshold + smoothness_i)^2
        end
      end
    end

    # normalize weights
    @views for k = 1:Npts
      if weno.requires_split[k]
        # eq after (2.4) is linear, so we can combine the coefficients here
        tmp_omegas[:,k] .= ( sigmas_plus[k] * omegas_plus[:,k] / sum(omegas_plus[:,k])
                            - sigmas_minus[k] * omegas_minus[:,k] / sum(omegas_minus[:,k]) )
      else
        tmp_omegas[:,k] .= omegas[:,k] / sum(omegas[:,k])
      end
    end

    fill!(tmp_uweno, 0.0)
    # reconstruct weno solution as a linear combination of the polynomials from different
    # stenciles with non-linear weights
    for i = 1:w+1
      subrange = 1+(i-1):1+(i-1)+w
      uavg_substencile = uavg_stencile[subrange] # covers ith reconstruction element
      p_i = transpose(uavg_substencile) * weno.subcell_polynomial_basis[i]
      @. tmp_uweno += tmp_omegas[i,:] * p_i[:]
    end

    # reinsert result
    mat_u[:,idx] .= tmp_uweno
  end

  return
end
