using InteractiveUtils
using PlotThemes


function main(parameters_filename="")

  parameters_filename = abspath(parameters_filename)
  outputdir = mk_outputdir(parameters_filename)
  cp(parameters_filename, joinpath(outputdir, basename(parameters_filename)))

  # read parameters
  if !isfile(parameters_filename)
    error("Cannot locate parameter file '$parameters_filename'")
  end
  prms = FlatDict(TOML.parsefile(parameters_filename))

  # setup RHS
  mesh      = make_Mesh(trim_headers(prms, "Mesh"))
  cache     = Cache(mesh)
  equation  = make_Equation(trim_headers(prms, "Equation"))
  tci       = TCI.make_TCI(trim_headers(prms, "TCI"), mesh)
  hrsc      = HRSC.make_HRSC(trim_headers(prms, "HRSC"), mesh)
  rsolver   = ApproxRiemannSolver(flux, speed, equation)
  ldg_rsolver, av_rsolver = if hrsc isa AbstractArtificialViscosity
    ApproxRiemannSolver(ldg_flux, ldg_speed, equation),
    ApproxRiemannSolver(av_flux, av_speed, equation)
  else
    nothing, nothing
  end
  rhs      = RHS(cache=cache, mesh=mesh, equation=equation, tci=tci, hrsc=hrsc,
                 rsolver=rsolver, ldg_rsolver=ldg_rsolver, av_rsolver=av_rsolver)

  _register_variables!(rhs)
  register_variables!(cache, rsolver)
  !isnothing(ldg_rsolver) && register_variables!(cache, ldg_rsolver)
  !isnothing(av_rsolver) && register_variables!(cache, av_rsolver, dont_register=true)
  display(cache)

  initialdata!(rhs, trim_headers(prms, "Initialdata", "Mesh"))

  default_boundaryconditions_prms = Dict("type"=>"periodic")
  boundaryconditions_prms = trim_headers(prms, "BoundaryConditions")
  amend!(boundaryconditions_prms, default_boundaryconditions_prms)
  diff = diffkeys(boundaryconditions_prms, default_boundaryconditions_prms)
  @assert length(diff) == 0 "Section 'BoundaryConditions': Unknown options '$diff' encountered!"
  bdryconds     = make_BoundaryConditions_equation(rhs, boundaryconditions_prms)
  ldg_bdryconds = make_BoundaryConditions_LDG(ldg_rsolver, boundaryconditions_prms)
  av_bdryconds  = make_BoundaryConditions_LDG(av_rsolver, boundaryconditions_prms)

  rhs_fn!(du, u, t) = rhs!(du, u, t, rhs, bdryconds, ldg_bdryconds, av_bdryconds)
  timestep_fn(u, t, n) = timestep(u, t, rhs)

  ### Callbacks
  cbs_fullstep, cbs_substep = make_callback(rhs, bdryconds)

  # Plotting
  default_plot_prms = Dict("every_iteration"=>0,
                           "every_dt"=>0.0,
                           "pause"=>false)
  plot_prms = trim_headers(prms, "Plot")
  dg1d.amend!(plot_prms, default_plot_prms)
  diff = diffkeys(plot_prms, default_plot_prms)
  theme(:dark, markerstrokewidth=0.1, marker=:circle)
  plotfn(u,t) = plot(rhs)
  @assert length(diff) == 0 "Section 'Plot': Unknown options '$diff' encountered!"
  plotcb = PlotCallback(plotfn,
                        CallbackTiming(every_dt=plot_prms["every_dt"],
                                       every_iteration=plot_prms["every_iteration"],
                                       label="Plotting"),
                        pause=plot_prms["pause"])
  default_output_prms = Dict("every_iteration"=>0,
                             "every_dt"=>0.5,
                             "variables"=>["u"],
                             "filename"=>joinpath(outputdir, "output"),
                             "aligned_times"=>[])
  output_prms = trim_headers(prms, "Output")
  amend!(output_prms, default_output_prms)
  diff = diffkeys(output_prms, default_output_prms)
  @assert length(diff) == 0 "Section 'Output': Unknown options '$diff' encountered!"
  savecb = SaveCallback(Symbol.(output_prms["variables"]), cache, mesh,
                        output_prms["filename"],
                        CallbackTiming(every_dt=output_prms["every_dt"],
                                       every_iteration=output_prms["every_iteration"],
                                       label="Save output")
                       )

  # artificially smooth initial data
  smoother_called = false
  function id_smoother_fn(u, t)
    if !smoother_called
      println("Smoothing initial data...")
      @unpack u     = get_dynamic_variables(cache)
      @unpack flag  = get_cell_variables(cache)
      id_hrsc = HRSC.BernsteinReconstruction(mesh)
      HRSC.reconstruct!(u, flag, id_hrsc, isperiodic=isperiodic(bdryconds))
      smoother_called = true
    end
  end
  id_smoother_cb = FunctionCallback(id_smoother_fn,
                                    CallbackTiming(every_iteration=1,
                                                   label="Initial data smoother"))
  append!(cbs_fullstep, [id_smoother_cb, savecb, plotcb])

  aligned_times = get(output_prms, "aligned_times", [])
  timealigned_savecb, timestep_fn = if length(aligned_times) > 0
    atscb = TimeAlignedSaveCallback(aligned_times, timestep_fn, savecb)
    atscb, atscb.aligned_timestep_fn
  else
    nothing, timestep_fn
  end

  !isnothing(timealigned_savecb) && push!(cbs_fullstep, timealigned_savecb)

  default_evolve_prms = Dict("algorithm"=>"LSERK4", "tend"=>1.0)
  evolve_prms = trim_headers(prms, "Evolution")
  amend!(evolve_prms, default_evolve_prms)
  diff = diffkeys(evolve_prms, default_evolve_prms)
  @assert length(diff) == 0 "Section 'Evolution': Unknown options '$diff' encountered!"

  @unpack u = get_dynamic_variables(cache)

  # display(cbs_fullstep)
  # return

  u = dg1d.evolve_ERK(rhs_fn!, u, timestep_fn,
                      (0.0, Float64(evolve_prms["tend"])),
                      Symbol(evolve_prms["algorithm"]),
                      callback_fullstep=cbs_fullstep,
                      callback_substep=cbs_substep)

  return
end
