# Alternative names:
# - Discretization ... too long
# - Approximation or Approx ... the second one might fit, but is also unspecific
# - Implementation or Impl ... too generic
# - Project ... too generic
# - State ... is at really a state?
Base.@kwdef mutable struct RHS{T_Equation <:AbstractScalarEquation, 
                               T_RSolver  <:AbstractRiemannSolver,
                               T_HRSC     <:Maybe{HRSC.AbstractHRSC},
                               T_TCI      <:Maybe{TCI.AbstractTCI},
                               T_LDG_RSolver<:Maybe{AbstractRiemannSolver},
                               T_AV_RSolver<:Maybe{AbstractRiemannSolver}}

  cache::Cache
  mesh::Mesh
  equation::T_Equation
  rsolver::T_RSolver
  hrsc::T_HRSC               = nothing
  tci::T_TCI                 = nothing
  ldg_rsolver::T_LDG_RSolver = nothing
  av_rsolver::T_AV_RSolver   = nothing
end


#######################################################################
#                         Register variables                          #
#######################################################################



# Julia limitation/bug: Overloading imported functions without their Module prefix shadows their
# definition in this module without a warning or error. This does not happen for methods
# defined in Base.
# Because of that we prefix with _ below.
# Cf. https://github.com/JuliaLang/julia/issues/15510
# Perhaps this is the import vs. using confusion people talk on the forums.
function _register_variables!(rhs::RHS)
  _register_variables!(rhs, rhs.equation)
  _register_variables!(rhs, rhs.hrsc)
  _register_variables!(rhs, rhs.tci)
end


function _register_variables!(rhs, eq::AbstractScalarEquation)
  register_variables!(rhs.cache,
    dynamic_variablenames = (:u,),
    rhs_variablenames     = (:rhs_u,),
    static_variablenames  = (:flx_u, :E, :Em1, :F, :Fm1, :v),
    cell_variablenames    = (:max_v,),
    global_variablenames  = (:t, :tm1))
end


function _register_variables!(rhs, hrsc::HRSC.AbstractArtificialViscosity)
  register_variables!(rhs.cache,
    static_variablenames  = (:q, :flx_q),
    cell_variablenames    = (:mu,))
end


function _register_variables!(rhs, hrsc::HRSC.SmoothedArtificialViscosity)
  _register_variables!(rhs, hrsc.av)
  register_variables!(rhs.cache,
    static_variablenames  = (:smoothed_mu,))
end


function _register_variables!(rhs, hrsc::Maybe{HRSC.AbstractHRSC}) end


function _register_variables!(rhs, tci::Nothing) end


function _register_variables!(rhs, tci::TCI.AbstractTCI)
  register_variables!(rhs.cache, cell_variablenames = (:flag,))
end


function _register_variables!(rhs, tci::TCI.EntropyProduction)
  register_variables!(rhs.cache, 
                      cell_variablenames   = (:flag,),
                      static_variablenames = (:EP,),
                      bdry_variablenames   = (:EPjump_lhs, :EPjump_rhs))
end


#######################################################################
#                              Timestep                               #
#######################################################################


function timestep(state_u, t, rhs::RHS)
  wrap_dynamic_variables!(rhs.cache, state_u)
  return timestep(t, rhs, rhs.hrsc)
end


function timestep(t, rhs::RHS, hrsc::Maybe{HRSC.AbstractHRSC})
  @unpack cache, mesh, equation = rhs
  @unpack v = get_static_variables(cache)

  broadcast_volume!(speed, equation, cache)
  vmax = dg1d.absolute_maximum(view(v,:))
  vmax_limit = 1e4
  if vmax > vmax_limit
    @warn "Limiting timestep due to maximum speed exceeding $vmax_limit"
    vmax = vmax_limit
  end

  @unpack dl, x, CFL, element = mesh
  @unpack N = element

  dt = CFL * dl / (N * vmax)
  return dt
end


#######################################################################
#                           RHS computation                           #
#######################################################################


function rhs!(state_du, state_u, t, rhs::RHS,
    bdryconds::BoundaryConditions,
    ldg_bdryconds::Maybe{BoundaryConditions},
    av_bdryconds::Maybe{BoundaryConditions})

  wrap_dynamic_variables!(rhs.cache, state_u)
  wrap_rhs_variables!(rhs.cache, state_du)
  return rhs!(rhs, t, rhs.hrsc, bdryconds, ldg_bdryconds, av_bdryconds)
end


function rhs!(rhs::RHS, t, hrsc::Nothing, bdryconds, ldg_bdryconds::Nothing, av_bdryconds::Nothing)

  @unpack cache, mesh, equation, rsolver = rhs

  broadcast_volume!(flux, equation, cache)
  broadcast_faces!(lax_friedrich_flux, rsolver, cache, mesh)
  broadcast_boundaryconditions!(central_flux, bdryconds, cache, mesh, t)

  @unpack flx_u, v               = get_static_variables(cache)
  @unpack u                      = get_dynamic_variables(cache)
  @unpack rhs_u                  = get_rhs_variables(cache)
  @unpack lhs_numflx_u, rhs_numflx_u = get_bdry_variables(cache)
  compute_rhs_weak_form!(rhs_u, flx_u, lhs_numflx_u, rhs_numflx_u, mesh)

  return
end


function rhs!(rhs::RHS, t, hrsc::Union{BernsteinReconstruction, WENOReconstruction},
    bdryconds, ldg_bdryconds::Nothing, av_bdryconds::Nothing)

  @unpack cache, mesh, equation, rsolver = rhs
  @unpack flag                           = get_cell_variables(cache)
  @unpack flx_u                          = get_static_variables(cache)
  @unpack u                              = get_dynamic_variables(cache)
  @unpack rhs_u                          = get_rhs_variables(cache)
  @unpack lhs_numflx_u, rhs_numflx_u     = get_bdry_variables(cache)

  broadcast_volume!(flux, equation, cache)
  broadcast_faces!(lax_friedrich_flux, rsolver, cache, mesh)
  broadcast_boundaryconditions!(lax_friedrich_flux, bdryconds, cache, mesh, t)

  compute_rhs_weak_form!(rhs_u, flx_u, lhs_numflx_u, rhs_numflx_u, mesh)

  return
end


function rhs!(rhs::RHS, t, hrsc::HRSC.AbstractArtificialViscosity,
    bdryconds, ldg_bdryconds, av_bdryconds)

  @unpack cache, mesh, equation, rsolver, ldg_rsolver, av_rsolver = rhs

  @unpack flx_u, flx_q, q, v         = get_static_variables(cache)
  @unpack u                          = get_dynamic_variables(cache)
  @unpack rhs_u                      = get_rhs_variables(cache)
  @unpack max_v                      = get_cell_variables(cache)
  @unpack lhs_numflx_u, rhs_numflx_u,
          lhs_numflx_q, rhs_numflx_q = get_bdry_variables(cache)

  ## solve auxiliary equation: q + ∂x u = 0
  broadcast_volume!(ldg_flux, equation, cache)
  broadcast_faces!(central_flux, ldg_rsolver, cache, mesh)
  broadcast_boundaryconditions!(central_flux, ldg_bdryconds, cache, mesh, t)
  compute_rhs_weak_form!(q, flx_q, lhs_numflx_q, rhs_numflx_q, mesh)

  ## compute rhs of regularized equation: ∂t u + ∂x f + ∂x μ q = 0
  # @unpack smoothed_mu = get_static_variables(cache)
  # @. flx_q = smoothed_mu * q
  broadcast_volume!(av_flux, equation, cache)
  broadcast_faces!(mean_flux, av_rsolver, cache, mesh)
  broadcast_boundaryconditions!(mean_flux, av_bdryconds, cache, mesh, t)
  broadcast_volume!(flux, equation, cache)
  broadcast_faces!(lax_friedrich_flux, rsolver, cache, mesh)
  broadcast_boundaryconditions!(lax_friedrich_flux, bdryconds, cache, mesh, t)

  # # add up fluxs and numerical fluxes
  @. flx_u += flx_q
  @. lhs_numflx_u += lhs_numflx_q
  @. rhs_numflx_u += rhs_numflx_q

  compute_rhs_weak_form!(rhs_u, flx_u, lhs_numflx_u, rhs_numflx_u, mesh)

  return
end
