Base.@kwdef struct Minmod <: AbstractTCI
  mesh::Mesh
  # parameters
  threshold::Float64
end


function parse_parameters(::Type{Minmod}, prms)
  threshold = query(Float64, prms, "threshold", default=50)
  @assert threshold > 0
  return (; threshold)
end


function compute_indicator!(
    flag, # mutated outputs
    u,    # inputs
    tci::Minmod)

  @unpack mesh       = tci
  @unpack invjac, dl = mesh
  @unpack D          = mesh.element
  L                  = layout(mesh)
  Npts, K            = L

  fill!(flag, 0)
  u = reshape(u, L)
  u_averages = dg1d.cellwise_average(u, mesh)

  @unpack threshold = tci
  # fill flag by applying minmod criterion
  for k in 1:K # loop over cells
    # assuming periodic grid
    k_lhs, k_rhs = periodic_index(k-1, K), periodic_index(k+1, K)
    uavg, uavg_lhs, uavg_rhs = u_averages[k], u_averages[k_lhs], u_averages[k_rhs]
    ubdry_lhs, ubdry_rhs = u[1,k], u[end,k]

    # definitions (17), (18) in Bugner et al. 2015 (paper)
    Δu_min = uavg - uavg_lhs
    Δu_max = uavg_rhs - uavg
    utilde_min = uavg - ubdry_lhs
    utilde_max = ubdry_rhs - uavg

    # troubled check, cf. (21) in Bugner et al. 2015 (paper), threshold =^= M
    if !(minmod_M(dl, threshold, utilde_min, Δu_min, Δu_max) ≈ utilde_min) ||
       !(minmod_M(dl, threshold, utilde_max, Δu_min, Δu_max) ≈ utilde_max)
      flag[k] = 1
    end
  end

  return
end


function minmod(args...)
  sign1 = args[1] > 0 ? 1.0 : -1.0
  result = abs(args[1])
  for i = 2:length(args)
    ai = args[i]
    sign1 * ai < 0.0 && return 0.0
    result = min(abs(ai), result)
  end
  return sign1 * result
end


function minmod(arg1, arg2)
  arg1 * arg2 < 0.0 && return 0.0
  return (arg1 > 0.0 ? 1.0 : -1.0) * min(abs(arg1), abs(arg2))
end


function minmod_M(dx::T, M::T, args::T...)::T where T <: Real
  if abs(args[1]) <= M * dx^2
    return args[1]
  else
    return minmod(args...)
  end
end
