#######################################################################
#                                Type                                 #
#######################################################################


struct _VariablesArray
  npts::Int64
  names::Vector{Symbol}
  data::Vector{Vector{Float64}}

  function _VariablesArray(npts, names, data)
    nvars = length(names)
    @assert length(data) == npts*nvars "Mismatch between number of points, number of variables " *
                                       "and length of data vector!"
    return new(npts, data, names)
  end
end


#######################################################################
#                            Constructors                             #
#######################################################################


function _VariablesArray(npts::Int64, vartpls::Tuple{Symbol, Vector{Float64}}...)
  names = [ vt[1] for vt in vartpls ]
  data  = [ vt[2] for vt in vartpls ]
  return _VariablesArray(npts, names, data)
end


function _VariablesArray(npts::Int64, varnames::Symbol...)
  varnames = [ vn for vn in varnames ]
  vardata = [ zeros(Float64, npts) for _ in 1:length(varnames) ]
  return _VariablesArray(npts, names, data)
end


_VariablesArray(npts::Int64) = _VariablesArray(npts, Symbol[], Vector{Float64}[])
_VariablesArray()            = _VariablesArray(0)


#######################################################################
#                               Methods                               #
#######################################################################


function add!(vars::_VariablesArray, vartpl::Tuple{Symbol, Vector{Float64}}, overwrite::Bool=false)
  varname, vardata = vartpl
  if !overwrite && varname in vars.names
    error("Variable $varname already present!")
  end
  @assert length(vardata) == vars.npts "Mismatch number of points and data vector length!"
  push!(vars.data, vardata)
  push!(vars.names, varname)

  if overwrite
    idx = findfirst(n -> n == varname, vars.names)
    if !isnothing(idx)
      vars.data[idx] = vardata
      vars.names[idx] = varname
    else
      push!(vars.data, vardata)
      push!(vars.names, varname)
    end
  else
    push!(vars.data, vardata)
    push!(vars.names, varname)
  end

  return
end


function findname(vars::_VariablesArray, name::Symbol)
  return findfirst(n -> n === name, vars.names)
end


function add!(vars::_VariablesArray, varnames::Symbol...)
  duplicated = [ varname in vars.names for varname in varnames ]
  if any(duplicated)
    idxs = [ findname(vars, varname) for varname in varnames ]
    idxsx = filter(!isnothing, idxs)
    duplicated_names = [ varnames[idx] for idx in idxs ]
    error("Variables '$duplicated_names' are already present.")
  end
  push!(vars.data, [ zeros(Float64, npts) for _ in 1:length(varnames) ]...)
  push!(vars.names, [ vn for vn in varnames ]... )
  return
end


function drop!(vars::_VariablesArray, varname::Symbol)
  idx = findfirst(n -> n == varname, vars.names)
  if !isnothing(idx)
    vardata = vars.data[idx]
    deleteat!(vars.data, idx)
    deleteat!(vars.names, idx)
    return vardata
  else
    error("Variable $varname not found!")
  end
end


function Base.getproperty(vars::_VariablesArray, sym::Symbol)
  names = getfield(vars, :names)
  idx = findfirst(n -> n == sym, names)
  if !isnothing(idx)
    return getfield(vars, :data)[idx]
  else
    return getfield(vars, sym)
  end
end


function Base.getproperty(vars::_VariablesArray, sym::String)
  return getproperty(vars, Symbol(sym))
end


function Base.propertynames(vars::_VariablesArray, private::Bool=false)
  return [ :data, :names, getfield(vars, :names)... ]
end


function Base.setproperty!(vars::_VariablesArray, sym::Symbol, var::Vector{Float64})
  names = getfield(vars, :names)
  idx = findname(vars, sym)
  if isnothing(idx)
    error("Variable '$sym' not present. You must first add it with add!")
  end
  data = getfield(vars, data)
  @assert length(data[idx]) == getfield(vars, :npts)
  data[idx] .= var
  return
end


n_variables(vars::_VariablesArray)  = length(vars.names)
Base.isinf(vars::_VariablesArray)   = any(isinf.(vars.data))
Base.isnan(vars::_VariablesArray)   = any(isnan.(vars.data))

## resize method?
# There are various implementations possible:
# - Extend resize vectors naively. This makes not much sense, because the data then has to be
# adjusted somewhere else
# - Provide an 'interpolateion/restriction interface', e.g. 
# resize(vars::_VariablesArray, resize_function) where the `resize_function` does interpolation
# or restriction in a clever way to avoid errors and keep data consistent
