@testset "RiemannSolver" begin


# setup a new equation type
struct MyDummyEquation <: AbstractEquation end

function register_variables!(cache::Cache, eq::MyDummyEquation)
  dg1d.register_variables!(cache, dynamic_variablenames = (:a, :b),
                             rhs_variablenames = (:f_a, :f_b),
                             cell_variablenames = (:max_v,))
end

# define methods needed for ApproxRiemannSolver interface

@with_signature function flux(::MyDummyEquation)
  @accepts State(a, b)
  f_a, f_b = a + b, a - b
  @returns f_a, f_b
end

@with_signature function speed(::MyDummyEquation)
  @accepts State(a, b)
  max_v = max(1.0 + 1.0, 1.0 - 1.0)
  @returns max_v
end

@with_signature function speed_no_state_tag(::MyDummyEquation)
  @accepts a, b
  max_v = max(1.0 + 1.0, 1.0 - 1.0)
  @returns max_v
end

@with_signature function speed_wrong_states(::MyDummyEquation)
  @accepts State(a, c)
  max_v = max(1.0 + 1.0, 1.0 - 1.0)
  @returns max_v
end


@testset "ApproxRiemannSolver" begin

  mesh = make_Mesh()
  cache = Cache(mesh)
  equation = MyDummyEquation()
  register_variables!(cache, equation)

  ### constructor
  rsolver = ApproxRiemannSolver(flux, speed, equation)

  ### check registration
  dg1d.register_variables!(cache, rsolver)
  to_be_registered_bdry_vars = [ :lhs_numf_a, :lhs_numf_b, :rhs_numf_a, :rhs_numf_b ]
  all_bdry_vars = dg1d.names(get_bdry_variables(cache))
  @test issubset(to_be_registered_bdry_vars, all_bdry_vars)

  ### only test if implemented methods compute something when used with
  ### broadcasting_cached!; actual result should be verified with integration tests

  # fill a, b with random numbers
  @unpack a, b = get_dynamic_variables(cache)
  map!.(el -> randn(), [a,b], [a,b])
  for fn in [ lax_friedrich_flux, central_flux, mean_flux ]
    broadcast_faces!(fn, rsolver, cache, mesh)
  end

  # mismatch between signatures of flux and speed functions
  @test_throws AssertionError ApproxRiemannSolver(flux, speed_wrong_states, equation)
  @test_throws AssertionError ApproxRiemannSolver(flux, speed_no_state_tag, equation)

end


@testset "BoundaryConditions" begin

  mesh = make_Mesh()
  cache = Cache(mesh)
  equation = MyDummyEquation()
  register_variables!(cache, equation)

  ### constructor
  rsolver = ApproxRiemannSolver(flux, speed, equation)
  dg1d.register_variables!(cache, rsolver)

  lhs_bc = DirichletBC(t -> (0.0,1.0))
  rhs_bc = OutflowBC()
  bdryconds = BoundaryConditions(lhs_bc, rhs_bc, rsolver)

  ### only test if implemented methods compute something when used with
  ### broadcasting_cached!; actual result should be verified with integration tests

  for fn in [ lax_friedrich_flux, central_flux ]
    dg1d.broadcast_boundaryconditions!(fn, bdryconds, cache, mesh, 0.0)
  end

  ### edge cases

  # incompatible bdry conitions
  @test_throws AssertionError BoundaryConditions(DirichletBC(t-> (0.0,)),
                                                 DirichletBC(t-> (0.0,1.0,2.0)),
                                                 rsolver)
  # mismatch between boundary states and flux arguments
  @test_throws AssertionError BoundaryConditions(DirichletBC(t-> (0.0,)),
                                                 DirichletBC(t-> (1.0,)),
                                                 rsolver)
  # single periodic boundary does not make sense
  @test_throws AssertionError BoundaryConditions(PeriodicBC(),
                                                 DirichletBC(t-> (1.0,)),
                                                 rsolver)

end


end
