@testset "@with_signature" begin

  struct DummyEquation end
  dummyeq = DummyEquation()

  # general usage
  @with_signature function fn1(eq::DummyEquation)
    @accepts u1, u2
    rhs_u1, rhs_u2 = 2.0 * u1 + u2, u2 * 3.0 - u1
    @returns rhs_u1, rhs_u2
  end
  @test accepts(fn1, dummyeq) == (:u1, :u2)
  @test returns(fn1, dummyeq) == (:rhs_u1, :rhs_u2)
  @test state_indices(fn1, dummyeq) == ()
  @test has_signature(fn1, dummyeq) == true
  args = (1.0, 2.0)
  obtained = fn1(args, dummyeq)
  expected = (4.0, 5.0)
  @test obtained == expected

  # @accepts and @returns with only one argument each
  @with_signature function fn2(eq::DummyEquation)
    @accepts u1
    rhs_u1 = 2.0 * u1
    @returns rhs_u1
  end
  @test accepts(fn2, dummyeq) == (:u1,)
  @test returns(fn2, dummyeq) == (:rhs_u1,)
  @test state_indices(fn2, dummyeq) == ()
  @test has_signature(fn2, dummyeq) == true
  args = (1.0,)
  obtained = fn2(args, dummyeq)
  expected = (2.0,)
  @test obtained == expected


  # @accepts State variables
  @with_signature function fn3(eq::DummyEquation)
    @accepts State(u1, u2)
    rhs_u1, rhs_u2 = 2.0 * u1 + u2, u2 * 3.0 - u1
    @returns rhs_u1, rhs_u2
  end
  @test accepts(fn3, dummyeq) == (:u1, :u2)
  @test returns(fn3, dummyeq) == (:rhs_u1, :rhs_u2)
  @test state_indices(fn3, dummyeq) == (1,2)
  @test has_signature(fn3, dummyeq) == true
  args = (1.0, 2.0)
  obtained = fn3(args, dummyeq)
  expected = (4.0, 5.0)
  @test obtained == expected

  # @accepts single State variable
  @with_signature function fn4(eq::DummyEquation)
    @accepts State(u1)
    rhs_u1 = 2.0 * u1
    @returns rhs_u1
  end
  @test accepts(fn4, dummyeq) == (:u1,)
  @test returns(fn4, dummyeq) == (:rhs_u1,)
  @test state_indices(fn4, dummyeq) == (1,)
  @test has_signature(fn4, dummyeq) == true
  args = (1.0,)
  obtained = fn4(args, dummyeq)
  expected = (2.0,)
  @test obtained == expected

  # @accepts mix of variables and State variables
  @with_signature function fn5(eq::DummyEquation)
    @accepts State(u1), u2, State(u3, u4)
    rhs_u1 = 2.0 * u1 + u2 + u3 + u4
    @returns rhs_u1
  end
  @test accepts(fn5, dummyeq) == (:u1,:u2,:u3,:u4)
  @test returns(fn5, dummyeq) == (:rhs_u1,)
  @test state_indices(fn5, dummyeq) == (1,3,4)
  @test has_signature(fn5, dummyeq) == true
  args = (1.0,2.0,3.0,4.0)
  obtained = fn5(args, dummyeq)
  expected = (11.0,)
  @test obtained == expected

  # no variable name for dispatch type required
  @with_signature function fn6(#= missing name here =#::DummyEquation)
    @accepts u1
    rhs_u1 = 2.0 * u1
    @returns rhs_u1
  end
  @test accepts(fn6, dummyeq) == (:u1,)
  @test returns(fn6, dummyeq) == (:rhs_u1,)
  @test state_indices(fn6, dummyeq) == ()
  @test has_signature(fn6, dummyeq) == true
  args = (1.0,)
  obtained = fn6(args, dummyeq)
  expected = (2.0,)
  @test obtained == expected

  # function with unused variable rhs_u2 in body throws normal UndefVarError
  @with_signature function fn_undefvar(eq::DummyEquation)
    @accepts u1
    rhs_u1 = 2.0 * u1
    @returns rhs_u1, rhs_u2
  end
  args = (1.0,)
  @test_throws UndefVarError fn_undefvar(args, dummyeq)

  # missing functiton body error
  @test_throws LoadError @eval @with_signature function fn(eq::DummyEquation) end

  # @accepts/@returns must appear at the start/end of the function body
  @test_throws LoadError @eval @with_signature function fn_w_error(eq::DummyEquation)
    println("this line is not allowed")
    @accepts u1
    rhs_u1 = 2.0 * u1
    @returns rhs_u1, rhs_u2
  end
  @test_throws LoadError @eval @with_signature function fn_w_error(eq::DummyEquation)
    @accepts u1
    rhs_u1 = 2.0 * u1
    @returns rhs_u1, rhs_u2
    println("this line is not allowed")
  end

  # @accepts/@returns can only appear once
  @test_throws LoadError @eval @with_signature function fn_w_error(eq::DummyEquation)
    @accepts u1
    @accepts u1
    rhs_u1 = 2.0 * u1
    @returns rhs_u1, rhs_u2
  end
  @test_throws LoadError @eval @with_signature function fn_w_error(eq::DummyEquation)
    @accepts u1
    rhs_u1 = 2.0 * u1
    @returns rhs_u1, rhs_u2
    @returns rhs_u1, rhs_u2
  end

  # @accepts and @returns require each at least one argument
  @test_throws LoadError @eval @with_signature function fn_w_error(eq::DummyEquation)
    @accepts
    rhs_u1 = 1
    @returns rhs_u1
  end
  @test_throws LoadError @eval @with_signature function fn_w_error(eq::DummyEquation)
    @accepts u1
    rhs_u1 = 2.0 * u1
    @returns
  end

  # function definition requires exactly one argument
  @test_throws LoadError @eval @with_signature function fn_w_error(#=no argument=#)
    @accepts u1
    rhs_u1 = 2.0 * u1
    @returns rhs_u1
  end
  @test_throws LoadError @eval @with_signature function fn_w_error(eq::DummyEquation, too_much)
    @accepts u1
    rhs_u1 = 2.0 * u1
    @returns rhs_u1
  end

  # the only argument must be type annotated
  @test_throws LoadError @eval @with_signature function fn_w_error(eq)
    @accepts u1
    rhs_u1 = 2.0 * u1
    @returns rhs_u1
  end

  # fallback interface methods error when called
  function fn_with_no_signature(eq::DummyEquation)
    @accepts u1
    rhs_u1 = 2.0 * u1
    @returns rhs_u1
  end
  @test_throws ErrorException accepts(fn_with_no_signature,dummyeq)
  @test_throws ErrorException returns(fn_with_no_signature,dummyeq)
  @test_throws ErrorException state_indices(fn_with_no_signature,dummyeq)
  # expect the trait function is just false
  has_signature(fn_with_no_signature, dummyeq) == false

end
